from fabric.api import *
from py_mina.decorators import task, deploy_task, setup_task
from py_mina.config import config, fetch, ensure, set
from py_mina.state import state
from py_mina.echo import echo_task, echo_subtask, echo_comment, echo_status



"""
Templates
"""


from py_mina.cli.template_generator import *


################################################################################
# Base
################################################################################


base_template = template_places_6.format(
    generate_head(),
    generate_settings('global'),
    generate_settings('connection'),
    generate_settings('application'),
    generate_settings('shared'),
    generate_tasks()
)


################################################################################
# Staged
################################################################################


def generate_staged_templates():
    return {
        'settings': generate_settings('global', staged=True),
        'tasks': template_places_3.format(
            generate_head(staged=True, import_subtasks=True), 
            generate_settings('shared', staged=True),
            generate_tasks(staged=True)
        ),
        'production': template_places_3.format(
            generate_head(staged=True),
            generate_settings('connection', staged='production'),
            generate_settings('application', staged='production'),
        ),
        'dev': template_places_3.format(
            generate_head(staged=True),
            generate_settings('connection', staged='dev'),
            generate_settings('application', staged='dev'),
        ),
    }



"""
Decorator for setup task
"""


from __future__ import with_statement
import timeit
from fabric.api import task, settings, show, hide
from py_mina.subtasks.setup import *
from py_mina.echo import echo_task, print_task_stats


def setup_task(wrapped_function):
	"""
	Setup task function decorator
	"""

	wrapped_function_name = wrapped_function.__name__

	def _pre_setup():
		"""
		Creates required structure and adds hosts to known hosts
		"""
		with settings(hide('output')):
			create_required_structure()				
			add_repo_to_known_hosts()
			add_host_to_known_hosts()


	def setup_wrapper(*args):
		"""
		Runs setup process on remote server
			1) check required settings (user, host, deploy_to)
			2) creates required structure (releases, shared, tmp)   
			3) adds repo and host to known hosts if possible
			4) runs wrpapped function
		"""

		check_setup_config()

		start_time = timeit.default_timer()

		echo_task('Running "%s" task' % wrapped_function_name)

		with settings(show('debug'), colorize_errors=True):
			try:
				_pre_setup()
				wrapped_function(*args)

				print_task_stats(wrapped_function_name, start_time)
			except Exception as e:
				print_task_stats(wrapped_function_name, start_time, e)

	# Copy __name__ and __doc__ from decorated function to decorator function
	setup_wrapper.__name__ = wrapped_function_name or 'setup'
	if wrapped_function.__doc__: setup_wrapper.__doc__ = wrapped_function.__doc__

	# Decorate with `fabric3` task decorator
	return task(setup_wrapper)



"""
Generates py_mina deployfile templates
"""


pymina_head = '''"""
py_mina deployfile
"""\n\n'''


################################################################################
# Settings
################################################################################


def generate_settings(settings_type='global', staged=False):
    if settings_type == 'global': return generate_settings_global(staged)
    if settings_type == 'application': return generate_settings_application(staged)
    if settings_type == 'shared': return generate_settings_shared(staged)
    if settings_type == 'connection': return generate_settings_connection(staged)


def generate_settings_global(staged=False):
    staged_import = ('%s\nfrom py_mina import set\n\n\n' % pymina_head) if staged != False else ''

    return '''{0}# Settings - global


set('verbose', True)
set('keep_releases', 5)
#set('sudo_on_chown', True)
#set('sudo_on_chmod', True)
#set('ask_unlock_if_locked', True)'''.format(staged_import)


def generate_settings_application(staged=False):
    branch = 'master'
    application_name = 'example_application'
    application_string = '%s\'' % application_name

    if staged != False:
        branch = 'master' if staged == 'production' else staged
        application_string = '{0}_{1}'.format(
            application_name, 
            '''%s' % fetch('stage_name')'''
        )

    return '''# Settings - application


set('deploy_to', '/var/www/{0})
set('repository', 'https://github.com/py-mina-deploy/dummy-web-for-deployment')
set('branch', '{1}')'''.format(application_string, branch)


def generate_settings_shared(staged=False):
    return '''# Settings - shared [PUBLIC] files/dirs (application configs, assets, storage, etc.)


set('shared_dirs', [])
set('shared_files', [])


# Settings - explicit owner of [PUBLIC] shared files/dirs


#set('owner_user', 'www-data')
#set('owner_group', 'www-data')


# Settings - protected shared files/dirs (db configs, certificates, keys, etc.)
#          * [PROTECTED] owner config settings are required to be set


#set('protected_shared_dirs', [])
#set('protected_shared_files', [])


# Settings - owner of [PROTECTED] shared files/dirs

#set('protected_owner_user', 'root')
#set('protected_owner_group', 'root')'''


def generate_settings_connection(staged=False):
    staged_name = ("\nset('stage_name', '%s')" % staged) if staged != False else ''

    return '''# Settings - remote server connection


set('user', 'root')
set('hosts', ['localhost']){0}'''.format(staged_name)


################################################################################
# Head and Tasks
################################################################################


def generate_head(staged=False, import_subtasks=False):
    subtasks = '''from py_mina.subtasks import (
    git_clone, 
    create_shared_paths, 
    link_shared_paths, 
    rollback_release, 
    force_unlock,
)'''
    staged_without_subtasks = '''{0}
from py_mina import set, fetch
from tasks import *
import settings'''.format(pymina_head)
    staged_with_subtasks = '''{0}
from py_mina import *
{1}'''.format(pymina_head, subtasks)
    staged_return = staged_with_subtasks if import_subtasks else staged_without_subtasks

    return staged_return if staged else '%s\nfrom py_mina import *\n%s' % (pymina_head, subtasks)



def generate_tasks(staged=False):
    return '''# Tasks


@task
def restart():
    """
    Restarts application on remote server
    """
    
    with cd(fetch('current_path')):
        run('touch restart.txt')


@deploy_task(on_success=restart)
def deploy():
    """
    Runs deploy process on remote server
    """

    git_clone()
    link_shared_paths()


@setup_task
def setup():
    """
    Runs setup process on remote server
    """

    create_shared_paths()


@task
def rollback():
    """
    Rollbacks to previous release
    """

    rollback_release()


@task
def unlock():
    """
    Forces lockfile removal when previous deploy failed
    """

    force_unlock()
'''


################################################################################
# Place template
################################################################################


template_places_2 = '''{0}


{1}'''


template_places_3 = '''{0}


{1}


{2}'''


template_places_4 = '''{0}


{1}


{2}


{3}'''


template_places_5 = '''{0}


{1}


{2}


{3}


{4}'''

template_places_6 = '''{0}


{1}


{2}


{3}


{4}


{5}'''



"""
Console output
"""


import os
import timeit
from fabric.colors import *


################################################################################
# Echo
################################################################################


# System


def echo_task(message, error=False):
    color = red if error == True else yellow
    
    print(color('\n=====> %s' % message))


def echo_subtask(message, error=False):
    color = red if error == True else cyan

    print(cyan('\n-----> %s' % message))


# For user


def echo_comment(message, error=False):
    color = red if error == True else blue

    print(color(message))


def echo_status(message, error=False):
    color = red if error == True else green

    print(color(message))


################################################################################
# Print
################################################################################


# Task


def print_task_stats(task_name, start_time, error=None):
    status_tuple = (task_name, time_string(start_time))

    if error:
        echo_comment(('\n[FATAL ERROR]\n\n%s' % error), error=True)
        echo_status(('\n=====> Task "%s" failed %s \n' % status_tuple), error=True)
    else:
        echo_status('\n=====> Task "%s" finished %s \n' % status_tuple)
    

# Helpers


def time_string(start_time):
    stop_time = timeit.default_timer()
    delta_time = stop_time - start_time

    return '(time: %s seconds)' % delta_time


# CLI


def cli_staged_init_warning(default_directory = 'deploy'):
    notifictaion = red('\n[WARNING]\n\nDon\'t forget to specify %s %s' % (white('deployfile FILEPATH (-f)'), red('when running staged deployment')))
    prod_path = white('-f %s' % os.path.join(default_directory, 'production.py'))
    dev_path = white('-f %s' % os.path.join(default_directory, 'dev.py'))

    staged_deploy_notification = '''{0}

Production:
    $ py_mina {1} run setup
    $ py_mina {1} run deploy

Dev:
    $ py_mina {2} run setup
    $ py_mina {2} run deploy'''.format(notifictaion, prod_path, dev_path)

    print(staged_deploy_notification)

    



"""
Setup tasks
"""


from __future__ import with_statement
import timeit
import os
import re
from py_mina.config import fetch, ensure, check_config
from fabric.api import run, sudo, settings, env, cd, hide
from py_mina.echo import *


################################################################################
# Check
################################################################################


def check_setup_config():
    """
    Checks required config settings for setup task
    """

    check_config(['user', 'hosts', 'deploy_to'])


################################################################################
# Create required and shared 
################################################################################


def create_required_structure():
    """
    Creates required folders (tmp, releases, shared) in `deploy_to` path
    """

    deploy_to = fetch('deploy_to')

    echo_subtask('Creating required structure')

    create_entity(deploy_to, entity_type='directory', protected=False)

    for required_path in ['shared', 'releases', 'tmp']:
        create_entity(
            '/'.join([deploy_to, required_path]), 
            entity_type='directory', 
            protected=False
        )



def create_shared_paths():
    """
    Creates shared (+protected) files/dirs and sets unix owner/mode
    """

    global shared_path
    shared_path = fetch('shared_path')

    def create_dirs(directories, protected=False):
        global shared_path

        for sdir in directories:
            create_entity(
                '/'.join([shared_path, sdir]), 
                entity_type='directory', 
                protected=protected
            )


    def create_files(files, protected=False):
        global shared_path

        for sfile in files:
            directory, filename_ = os.path.split(sfile)

            if directory: 
                create_entity(
                    '/'.join([shared_path, directory]), 
                    entity_type='directory', 
                    protected=protected
                )

            filepath = '/'.join([shared_path, sfile])
            create_entity(filepath, entity_type='file', protected=protected)

            recommendation_tuple = (env.host_string, filepath)
            echo_status('\n=====> Don\'t forget to update shared file:\n[%s] %s\n' % recommendation_tuple, error=True)


    echo_subtask('Creating shared paths')

    # Shared
    create_dirs(fetch('shared_dirs', default_value=[]), protected=False)
    create_files(fetch('shared_files', default_value=[]), protected=False)

    # Protected
    create_dirs(fetch('protected_shared_dirs', default_value=[]), protected=True)
    create_files(fetch('protected_shared_files', default_value=[]), protected=True)


def create_entity(entity_path, entity_type = 'file', protected=False):
    """
    Creates directory/file and sets proper owner/mode.
    """

    chmod_sudo = 'sudo ' if fetch('sudo_on_chmod') else ''
    chown_sudo = 'sudo ' if fetch('sudo_on_chown') else ''

    def change_owner(owner_triple):
        """
        Changes unix owner/mode
        """

        run(chmod_sudo + 'chmod 0755 ' + entity_path)
        run(chown_sudo + 'chown %s:%s %s' % owner_triple)


    # 1) Create file/directory

    if entity_type == 'file':
        run('touch %s' % entity_path)
    else:
        run('mkdir -p %s' % entity_path)

    # 2) Change owner/mode

    if protected:
        protected_owner_user = fetch('protected_owner_user')
        protected_owner_group = fetch('protected_owner_group')

        change_owner((protected_owner_user, protected_owner_group, entity_path))
    else:
        owner_user = fetch('owner_user', default_value=False)
        owner_group = fetch('owner_group', default_value=False)
        
        if owner_user != False and owner_group != False: 
            change_owner((owner_user, owner_group, entity_path))
        else:
            run(chmod_sudo + 'chmod 0755 ' + entity_path)


################################################################################
# SSH known hosts
################################################################################


def add_repo_to_known_hosts():
    """
    Adds repository host to known hosts if possible
    """

    maybe_repository = fetch('repository', default_value='')

    if maybe_repository:
        # Parse repo host
        repo_host_array = re.compile('(@|://)').split(maybe_repository)
        repo_host_part = repo_host_array[-1] if repo_host_array else ''
        repo_host_match = re.compile(':|\/').split(repo_host_part)
        repo_host = repo_host_match[0] if repo_host_match else ''

        # Exit if no host
        if repo_host == '': 
            return

        # Parse port
        repo_port_match = re.search(r':([0-9]+)', maybe_repository)
        repo_port = repo_port_match.group(1) if bool(repo_port_match) else 22

        echo_subtask('Adding repository to known hosts')

        add_to_known_hosts(repo_host, repo_port)


def add_host_to_known_hosts():
    """
    Adds current host to the known hosts

    `fabric3` library sets to `env.host_string` current host where task is executed
    """

    ensure('hosts')

    echo_subtask('Adding current host to known hosts')

    add_to_known_hosts(env.host_string, env.port)


def add_to_known_hosts(host, port):
    cmd = '''
if ! ssh-keygen -H -F {0} &>/dev/null; then
    ssh-keyscan -t rsa -p {1} -H {0} >> ~/.ssh/known_hosts
fi'''.format(host, port)

    with settings(hide('everything'), warn_only=True):
        run(cmd)



"""
py_mina CLI commands
"""


import os
import time
from fabric.main import load_fabfile
from fabric.tasks import execute
from fabric.colors import green, red, yellow
from fabric.contrib.console import confirm
from py_mina.cli.templates import base_template, generate_staged_templates
from py_mina.echo import cli_staged_init_warning


#
# List
#


def cli_command_list(args):
  """
  Runs task from deploy file
  """
  
  filepath = args.get('filename')
  task_name = args.get('task')

  if filepath:
    fullpath = os.path.join(os.path.realpath(os.curdir), filepath)
    docstring, callables, default = load_fabfile(fullpath)

    if not docstring == None:
      print(docstring.rstrip('\n'))

    print('\nAvailable commands:\n')

    for task_name in callables:
      task_desc = callables.get(task_name).__doc__

      if not task_desc == None:
        description = task_desc.lstrip('\n').rstrip('\n')
      else:
        description = '\n'

      print('%s   %s' % (green(task_name), description))


#
# Run
#


def cli_command_run(args):
  """
  Runs task from deploy file
  """
  
  filepath = args.get('filename')
  task_name = args.get('task')

  if filepath:
    fullpath = os.path.join(os.path.realpath(os.curdir), filepath)
    docstring, callables, default = load_fabfile(fullpath)

    if task_name in callables:
      execute(callables[task_name])


#
# Init
#


def cli_command_init(args):
  """
  Creates init files (default folder path is `deploy/`)
  """
  
  # Show backup message after finish?
  global backup_happened

  def backup_extension(timestamp):
    """
    Generates extenion for backup file
    """

    return str(timestamp) + '.backup'


  def create_or_backup(filepath, content, backup_timestamp = None):
    """
    Creates new file and backups old file if necessary
    """

    if filepath:
      fullpath = os.path.join(os.path.realpath(os.curdir), filepath)
      directory, deployfile = os.path.split(fullpath)
      timestamp = time.time() if backup_timestamp == None else backup_timestamp
      backup_fullpath = "%s.%s" % (fullpath, backup_extension(timestamp))

      try:
        # deployfile directory (create)
        if not os.path.exists(directory):
          os.makedirs(directory)

        # deployfile (backup, create)
        if not os.path.exists(fullpath):
          create_file(filepath, content)
        else:
          if confirm('File %s exists. Backup?' % red(filepath)) == True:
            global backup_happened
            backup_happened = True

            os.rename(fullpath, backup_fullpath)
            print(green('Created backup file %s' % yellow(backup_fullpath)))

          create_file(filepath, content)

      except Exception as error: 
        print(red(error))


  def create_file(filepath, content):
    """
    Writes content to file
    """

    fullpath = os.path.join(os.path.realpath(os.curdir), filepath)

    try:
      with open(filepath, 'w') as f: f.write(content)
      print(green('Deployfile successfully created in %s' % yellow(fullpath)))
    except Exception as error: 
      print(red(error))


  backup_happened = False
  staged = args.get('staged')
  backup_timestamp = time.time()

  if staged == False:
    create_or_backup(args.get('filename'), base_template, backup_timestamp)
  else:
    default_directory = 'deploy/' if staged == None else staged
    templates_files = generate_staged_templates()

    for template_file_key in templates_files.keys():
      filepath = os.path.join(default_directory, '%s.py' % template_file_key)
      create_or_backup(filepath, templates_files[template_file_key], backup_timestamp)

    cli_staged_init_warning(default_directory)

  # show backup message
  if backup_happened:
    print(green('All backuped files has extension ') + red(backup_extension(backup_timestamp)))

